"""
Hourly Class

Retrieve hourly weather observations for one or multiple weather stations

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

import os
import datetime
import pandas as pd
from meteostat.core import Core
from math import floor, nan
from copy import copy

class Hourly(Core):

  # The list of weather Stations
  _stations = None

  # The start date
  _start = None

  # The end date
  _end = None

  # The data frame
  _data = pd.DataFrame(index = ['station', 'time'])

  # Columns
  _columns = [
    'date',
    'hour',
    'temp',
    'dwpt',
    'rhum',
    'prcp',
    'snow',
    'wdir',
    'wspd',
    'wpgt',
    'pres',
    'tsun',
    'coco'
  ]

  # Columns for date parsing
  _parse_dates = { 'time': [0, 1] }

  # Default aggregation functions
  _aggregations = {
    'time': 'first',
    'temp': 'mean',
    'dwpt': 'mean',
    'rhum': 'mean',
    'prcp': 'sum',
    'snow': 'mean',
    'wdir': 'mean',
    'wspd': 'mean',
    'wpgt': 'max',
    'pres': 'mean',
    'tsun': 'sum',
    'coco': 'max'
  }

  def _get_data(self, stations = None):

      if len(stations.index) > 0:

          paths = []

          for index, row in stations.iterrows():
              paths.append('hourly/' + row['id'] + '.csv.gz')

          files = self._load(paths)

          if len(files) > 0:

              for file in files:

                  if os.path.isfile(file['path']) and os.path.getsize(file['path']) > 0:
                      df = pd.read_feather(file['path'])

                      self._data = self._data.append(df[(df['time'] >= self._start) & (df['time'] <= self._end)])

  def __init__(
    self,
    stations = None,
    start = None,
    end = None,
    cache_dir = None,
    max_age = None,
    max_threads = None
  ):

      # Configuration - Cache directory
      if cache_dir != None:
        self._cache_dir = cache_dir

      # Configuration - Maximum file age
      if max_age != None:
        self._max_age = max_age

      # Configuration - Maximum number of threads
      if max_threads != None:
          self._max_threads = max_threads

      # Set list of weather stations
      if isinstance(stations, pd.DataFrame):
          self._stations = stations
      else:
          self._stations = pd.DataFrame(stations, columns = ['id'])

      # Set start date
      self._start = start

      # Set end date
      self._end = end

      # Get data
      try:
          self._get_data(self._stations)
      except:
          raise Exception('Cannot read hourly data')

  def normalize(self):

      # Create temporal instance
      self._temp = copy(self)

      # List of columns
      columns = ['station', 'time']

      # Dynamically append columns
      for column in self._temp._columns[2:]:
          columns.append(column)

      # Create result DataFrame
      result = pd.DataFrame(columns = columns)

      # Go through list of weather stations
      for station in self._temp._stations['id'].tolist():
          # Create data frame
          df = pd.DataFrame(columns = columns)
          # Add time series
          df['time'] = pd.date_range(self._temp._start, self._temp._end, freq='1H')
          # Add station ID
          df['station'] = station
          # Add columns
          for column in columns[2:]:
              # Add column to DataFrame
              df[column] = nan

          result = pd.concat([result, df], axis = 0)

      # Merge data
      self._temp._data = pd.concat([self._temp._data, result], axis = 0).groupby(['station', 'time'], as_index = False).first()

      # Return class instance
      try:
          return self._temp
      finally:
          self._temp = None

  def interpolate(self, limit = 3):

      # Create temporal instance
      self._temp = copy(self)

      # Apply interpolation
      self._temp._data = self._temp._data.groupby('station').apply(lambda group: group.interpolate(method = 'linear', limit = limit, limit_direction = 'both', axis = 0))

      # Return class instance
      try:
          return self._temp
      finally:
          self._temp = None

  def aggregate(self, freq = None, functions = None, spatial = False):

      # Create temporal instance
      self._temp = copy(self)

      # Update default aggregations
      if functions is not None:
          self._temp._aggregations.update(functions)

      # Time aggregation
      self._temp._data = self._temp._data.groupby(['station', pd.Grouper(key = 'time', freq = freq)]).agg(self._temp._aggregations)

      # Spatial aggregation
      if spatial:
          self._temp._data = self._temp._data.groupby([pd.Grouper(key = 'time', freq = freq)]).mean()

      # Return class instance
      try:
          return self._temp
      finally:
          self._temp = None

  def coverage(self, parameter = None):

      expect = floor((self._end - self._start).total_seconds() / 3600) + 1

      if parameter == None:
          return len(self._data.index) / expect
      else:
          return self._data[parameter].count() / expect

  def count(self):

      # Return number of rows in DataFrame
      return len(self._data.index)

  def fetch(self):

      # Return data frame
      return copy(self._data)



"""
A Python library for accessing open weather and climate data

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

__appname__ = "meteostat"
__version__ = "0.2.0"

from .core import Core
from .stations import Stations
from .hourly import Hourly
from .daily import Daily



"""
Core Class

Base class that provides methods which are used across the package

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

import os
import time
import hashlib
import pandas as pd
from copy import copy
from multiprocessing.pool import ThreadPool

class Core:

  # Temporary class storage
  _temp = None

  # Base URL of the Meteostat bulk data interface
  _endpoint = 'https://bulk.meteostat.net/'

  # Location of the cache directory
  _cache_dir = os.path.expanduser('~') + os.sep + '.meteostat' + os.sep + 'cache'

  # Maximum age of a cached file in seconds
  _max_age = 24 * 60 * 60

  # Maximum number of threads used for downloading files
  _max_threads = 1

  def _get_file_path(self, path = False):

      if path:
          # Get file ID
          file_id = hashlib.md5(path.encode('utf-8')).hexdigest()
          # Return path
          return self._cache_dir + os.sep + file_id
      else:
          # Return false
          return False

  def _file_in_cache(self, file_path = False):

      # Make sure the cache directory exists
      if not os.path.exists(self._cache_dir):
          try:
              os.makedirs(self._cache_dir)
          except:
              raise Exception('Cannot create cache directory')

      if file_path:
          # Return the file path if it exists
          if os.path.isfile(file_path) and time.time() - os.path.getmtime(file_path) <= self._max_age:
              return True
          else:
              return False
      else:
          return False

  def _download_file(self, path = None):

      if path:
          # Get local file path
          local_path = self._get_file_path(path)

          # Check if file in cache
          if not self._file_in_cache(local_path):

              if path[-6:-3] == 'csv':

                  # Read CSV file from Meteostat endpoint
                  try:
                      df = pd.read_csv(self._endpoint + path, compression = 'gzip', names = self._columns, parse_dates = self._parse_dates)
                  except:
                      return False

                  # Set weather station ID
                  if self.__class__.__name__ == 'Hourly' or self.__class__.__name__ == 'Daily':
                      df['station'] = path[-12:-7]

              # Save as Feather
              df.to_feather(local_path)

          return {
              'path': local_path,
              'origin': path
          }

  def _load(self, paths = None):

      if paths:

          # Create array of local file paths
          files = []

          # Single-thread processing
          if self._max_threads < 2:

              for path in paths:
                  files.append(self._download_file(path))

          # Multi-thread processing
          else:

              try:
                  pool = ThreadPool(self._max_threads).imap_unordered(self._download_file, paths)
              except:
                  raise Exception('Cannot create ThreadPool')

              for file in pool:
                if file != False:
                    files.append(file)

          # Return list of local file paths
          return files

  def clear_cache(self, max_age = None):

      try:
          # Set max_age
          if max_age is None:
              max_age = self._max_age

          # Get current time
          now = time.time()

          # Go through all files
          for file in os.listdir(self._cache_dir):

              # Get full path
              path = os.path.join(self._cache_dir, file)

              # Check if file is older than max_age
              if now - os.path.getmtime(path) > max_age and os.path.isfile(path):

                  # Delete file
                  os.remove(path)

      except:
          raise Exception('Cannot clear cache')



from os import path
from setuptools import setup, find_packages

# Content of the README file
here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.md')) as f:
    long_description = f.read()

# Setup
setup(
     name = 'meteostat',
     version = '0.2.0',
     author = 'Meteostat',
     author_email = 'info@meteostat.net',
     description = 'Access and analyze historical weather and climate data with Python.',
     long_description = long_description,
     long_description_content_type = 'text/markdown',
     url = 'https://github.com/meteostat/meteostat-python',
     packages = find_packages(),
     include_package_data = True,
     install_requires = ['pandas', 'pyarrow'],
     license = 'MIT',
     classifiers = [
         'Programming Language :: Python :: 3',
         'License :: OSI Approved :: MIT License',
         'Operating System :: OS Independent',
     ],
 )



"""
Stations Class

Select weather stations from the full list of stations

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution-NonCommercial
4.0 International Public License.

The code is licensed under the MIT license.
"""

import pandas as pd
from meteostat.core import Core
from math import cos, sqrt, radians
from copy import copy

class Stations(Core):

  # The list of selected weather Stations
  _stations = None

  # Columns
  _columns = [
    'id',
    'name',
    'country',
    'region',
    'wmo',
    'icao',
    'latitude',
    'longitude',
    'elevation',
    'timezone',
    'hourly_start',
    'hourly_end',
    'daily_start',
    'daily_end'
  ]

  # Columns for date parsing
  _parse_dates = [10, 11, 12, 13]

  def __init__(
    self,
    lat = None,
    lon = None,
    radius = None,
    country = None,
    region = None,
    bounds = None,
    id = None,
    wmo = None,
    icao = None,
    daily = None,
    hourly = None,
    cache_dir = None,
    max_age = None,
    max_threads = None
  ):

      # Configuration - Cache directory
      if cache_dir != None:
        self._cache_dir = cache_dir

      # Configuration - Maximum file age
      if max_age != None:
        self._max_age = max_age

      # Configuration - Maximum number of threads
      if max_threads != None:
          self._max_threads = max_threads

      # Get all weather stations
      try:
          file = self._load(['stations/lib.csv.gz'])[0]
          self._stations = pd.read_feather(file['path'])
      except:
           raise Exception('Cannot read weather station directory')

      # Filter by identifier
      if id != None or wmo != None or icao != None:
          self._identifier(id, wmo, icao)

      # Filter by country or region
      if country != None or region != None:
          self._regional(country, region)

      # Filter by boundaries
      if bounds != None:
          self._area(bounds)

      # Filter by distance
      if lat != None and lon != None:
          self._nearby(lat, lon, radius)

      # Filter by daily inventory
      if daily != None:
          self._inventory(daily, None)

      # Filter by hourly inventory
      if hourly != None:
          self._inventory(None, hourly)

  def _identifier(self, id = None, wmo = None, icao = None):

      # Get station by Meteostat ID
      if id != None:

          if not isinstance(id, list):
              id = [id]

          self._stations = self._stations[self._stations['id'].isin(id)]

      # Get station by WMO ID
      elif wmo != None:

          if not isinstance(wmo, list):
              wmo = [wmo]

          self._stations = self._stations[self._stations['wmo'].isin(wmo)]

      # Get stations by ICAO ID
      elif icao != None:

          if isinstance(icao, list):
              icao = [icao]

          self._stations = self._stations[self._stations['icao'] == icao]

      # Return self
      return self

  def _distance(self, station, point):

      # Earth radius in m
      R = 6371000

      x = (radians(point[1]) - radians(station['longitude'])) * cos(0.5 * (radians(point[0]) + radians(station['latitude'])))
      y = (radians(point[0]) - radians(station['latitude']))

      return R * sqrt(x * x + y * y)

  def _nearby(self, lat = False, lon = False, radius = None):

      # Get distance for each stationsd
      self._stations['distance'] = self._stations.apply(lambda station: self._distance(station, [lat, lon]), axis = 1)

      # Filter by radius
      if radius != None:
          self._stations = self._stations[self._stations['distance'] <= radius]

      # Sort stations by distance
      self._stations.columns.str.strip()
      self._stations = self._stations.sort_values('distance')

      # Return self
      return self

  def _regional(self, country = None, region = None):

      # Check if country is set
      if country != None:
          self._stations = self._stations[self._stations['country'] == country]

      # Check if region is set
      if region != None:
          self._stations = self._stations[self._stations['region'] == region]

      # Return self
      return self

  def _area(self, bounds = None):

      # Return stations in boundaries
      if bounds != None:
          self._stations = self._stations[(self._stations['latitude'] <= bounds[0]) & (self._stations['latitude'] >= bounds[2]) & (self._stations['longitude'] <= bounds[3]) & (self._stations['longitude'] >= bounds[1])]

      # Return self
      return self

  def _inventory(self, daily = None, hourly = None):

      # Check if daily is set
      if daily != None:
          self._stations = self._stations[(self._stations['daily_start'] != None) & (self._stations['daily_start'] <= daily) & (self._stations['daily_end'] >= daily)]

      # Check if hourly is set
      if hourly != None:
          self._stations = self._stations[(self._stations['hourly_start'] != None) & (self._stations['hourly_start'] <= hourly) & (self._stations['hourly_end'] >= hourly)]

      return self

  def sample(self, limit = 1):

      # Create temporal instance
      self._temp = copy(self)

      # Randomize the order of weather stations
      self._temp._stations = self._temp._stations.sample(limit)

      # Return class instance
      try:
          return self._temp
      finally:
          self._temp = None

  def count(self):

      # Return number of weather stations in current selection
      return len(self._stations.index)

  def fetch(self, limit = False):

      if limit:
          # Return data frame with limit
          return copy(self._stations.head(limit))
      else:
          # Return all entries
          return copy(self._stations)
